import time
import unittest
import sys

if sys.version_info >= (3, 3):
    from unittest import mock
else:
    import mock
from houston.plugin.gcp import GCPHouston
from tests.test_houston import MockResponse


class MockFuture:
    """Mock PubSub future object"""

    @staticmethod
    def result():
        time.sleep(0.1)
        return


class MockPubSubResponse:
    """Mock GCP PubSub response"""

    @staticmethod
    def publish(topic, data, **kwargs):
        return MockFuture


class TestCallStageViaPubSub(unittest.TestCase):

    test_plan_description = {
        "name": "test",
        "stages": [
            {"name": "start", "downstream": ["end"], "params": {"table": "test.sql"}},
            {"name": "end", "params": {"table": "test.sql", "psq": "tp-test-end"}},
        ],
    }

    def test_call_stage_via_pubsub(self):
        with mock.patch(
            "houston.gcp.client.pubsub_v1.PublisherClient"
        ) as pubsub_client:
            with mock.patch("houston.interface.requests.request") as http:
                http.return_value = MockResponse(
                    status_code=200,
                    json_data={
                        "success": True,
                        "complete": False,
                        "next": ["end"],
                        "params": {"end": {"table": "test.sql", "psq": "tp-test-end"}},
                    },
                )
                houston = GCPHouston(
                    api_key="test-key", plan=self.test_plan_description
                )
                response = houston.end_stage("start", "test-launch-id")
                pubsub_client.return_value = MockPubSubResponse
                houston.project = "test-gcp-project"
                houston.trigger_all(response['next'], "test-launch-id")

    def test_pubsub_trigger(self):
        with mock.patch(
            "houston.gcp.client.pubsub_v1.PublisherClient"
        ) as pubsub_client:
            with mock.patch("houston.interface.requests.request") as http:
                http.return_value = MockResponse(
                    status_code=200,
                    json_data={
                        "success": True,
                        "complete": False,
                        "next": ["end"],
                        "params": {"end": {"table": "test.sql", "psq": "tp-test-end"}},
                    },
                )
                houston = GCPHouston(
                    api_key="test-key", plan=self.test_plan_description
                )
                response = houston.end_stage("start", "test-launch-id")
                pubsub_client.return_value = MockPubSubResponse
                houston.project = "test-gcp-project"

                for next_stage in response['next']:
                    topic = houston.get_params(next_stage)['psq']
                    houston.pubsub_trigger({'stage': next_stage, 'mission_id': "test-launch-id"}, topic)


if __name__ == "__main__":
    unittest.main()



"""Houston Utilities for Microsoft Azure

Recommended usage:

    # end stage and trigger downstream stages

    res = h.end_stage(stage)

    for next_stage in res['next']:

        topic = (your method of getting the topic for the 'next_stage',
                 e.g. stage_parameter, simple conversion function, hard coded map, etc.)

        topic_key = (your method of getting the topic key for the 'next_stage',
                     e.g. environment variable, key vault secret, load from text file, etc.)

        h.event_grid_trigger({"stage": next_stage, "mission_id": mission_id}, topic, topic_key)

"""

from houston.client import Houston
from houston.exceptions import HoustonServerError, HoustonServerBusy
from datetime import datetime, timezone
from azure.eventgrid import EventGridClient
from msrest.authentication import TopicCredentials
from msrest.exceptions import HttpOperationError
from retry import retry
import uuid


retry_wrapper = retry((HoustonServerError, HoustonServerBusy, OSError, HttpOperationError), tries=3, delay=1, backoff=100)


@retry_wrapper
def event_grid_trigger(client, data):
    if 'plan' not in data:
        data['plan'] = client.plan['name']

    # TODO: get topic from key vault

    stage = data.get('stage')
    if stage is None:
        raise ValueError("Cannot trigger stage: A stage name was not provided.")

    service = client.get_service_from_stage_name(stage)
    if service is None:
        raise ValueError(f"Cannot trigger stage '{stage}': no service is defined for this stage. "
                         f"See https://callhouston.io/docs#services.")

    trigger = service.get('trigger')
    topic = trigger.get('topic')
    topic_key = trigger.get('topic_key')

    publish_event_grid_event(data, topic, topic_key)


class AzureHouston(Houston):

    def event_grid_trigger(self, data, topic, topic_key):
        """Sends a message to the provided Event Grid topic with the provided data payload.

        :param dict data: content of the message to be sent. Should contain 'stage' and 'mission_id'. Can contain any
                          additional JSON serializable information.
        :param string topic: The host name of the topic, e.g. 'topic1.westus2-1.eventgrid.azure.net'
        :param string topic_key: A 44 character access key for the topic.
        """
        event_grid_trigger(self, data, topic, topic_key)


def publish_event_grid_event(data, topic_hostname, topic_key):
    """Sends a message to the provided Event Grid topic with the provided data payload.
    :param dict data: content of the message to be sent.
    :param string topic_hostname: The host name of the topic, e.g. 'topic1.westus2-1.eventgrid.azure.net'
    :param string topic_key: A 44 character access key for the topic.
    """
    credentials = TopicCredentials(topic_key)

    event_grid_client = EventGridClient(credentials)

    event_id = str(uuid.uuid4())

    event_grid_client.publish_events(
        topic_hostname,
        events=[{
            'id': event_id,
            'subject': "Houston Stage Trigger",
            'data': data,
            'event_type': 'HoustonStageTrigger',
            'event_time': datetime.now(timezone.utc),
            'data_version': 1
        }]
    )



"""Commands are additional high-level methods to allow users or Houston integrated services to carry out common tasks
with a single command. A Houston integrated service will a command when triggered with a message containing the
'command' attribute.
"""

import os
import logging
from functools import wraps

from retry import retry
from houston import Houston
from houston.exceptions import HoustonException, HoustonClientError, HoustonPlanNotFound
from typing import *
import time
import json

HOUSTON_MAX_WAIT_INVOCATIONS = os.getenv('HOUSTON_MAX_WAIT_INVOCATIONS', 150)
log = logging.getLogger(os.getenv('HOUSTON_LOG_NAME', "houston"))
retry_wrapper = retry((OSError, AttributeError), tries=3, backoff=3, delay=3, logger=log)


def init_client(func):
    """Initialise Houston client if not defined in wrapped function arguments."""
    @wraps(func)
    def inner(plan: str = None, client: Houston = None, *args, **kwargs):

        if plan is None:
            if client is None:
                raise ValueError("Either plan or client must be provided to run a command.")
            else:
                plan = client.plan['name']

        try:

            if client is None:
                try:
                    from houston.plugin.gcp import GCPHouston
                    client = GCPHouston(plan)
                except ImportError:
                    pass

            if client is None:
                try:
                    from houston.plugin.azure import AzureHouston
                    client = AzureHouston(plan)
                except ImportError:
                    client = Houston(plan)

        except HoustonPlanNotFound:
            if func.__name__ == "delete":
                return  # if the plan was already deleted then the client can't be initialised
            else:
                raise

        res = func(plan, client, *args, **kwargs)
        return res
    return inner


@init_client
def start(plan: str, client: Houston, stages: Union[str, List[str]] = None, ignore: Union[str, List[str]] = None,
          stage=None, **kwargs) -> bool:
    """Start a new mission. Creates a new mission ID and then trigger the first stages or the requested stages.
    If the requested stages are not the first stages their upstream dependencies will be ignored.
    """

    mission_id = client.create_mission()

    if ignore is not None:
        if isinstance(ignore, str):
            ignore = [a.strip() for a in ignore.split(",")]
        for s in ignore:
            try:
                client.ignore_stage(s, mission_id)
            except HoustonException:
                pass

    if stages is not None or stage is not None:
        starting_stages = stages if stages is not None else stage
        if isinstance(starting_stages, str):
            starting_stages = [a.strip() for a in starting_stages.split(",")]
    else:  # if not stages - determine first stages
        starting_stages = [s['name'] for s in client.independent_stages]

    # trigger with dependencies ignored
    client.trigger_all(starting_stages, mission_id=mission_id, ignore_dependencies=True)

    log.info(f"Started new mission with stage{'s' if len(starting_stages) != 1 else ''}: {starting_stages}.")
    return True  # end


@init_client
def ignore(plan: str, client: Houston, mission_id: str, stages: Union[str, List[str]] = None,
           stage=None, **kwargs) -> bool:
    """Ignore the requested stages. If no stages are specified then every stage will be ignored (essentially stopping
    the mission. note: Houston cannot stop a stage that has already been started).
    """
    if stage is not None:
        stages = stage
    if stages is not None:
        stages = stages if isinstance(stages, list) else [a.strip() for a in stages.split(",")]
    else:
        stages = [s['name'] for s in client.plan['stages']]

    for s in stages:
        try:
            client.ignore_stage(s, mission_id)
        except HoustonException:
            pass  # we don't care if stage was already ignored
    log.info(f"Ignored stages: {', '.join(stages)}")
    return True  # end


@init_client
def static_fire(plan: str, client: Houston, stage: str, **kwargs) -> bool:
    """Run requested stage and in isolation; ignore dependencies and dependants.
    """
    mission_id = client.create_mission()

    client.trigger(dict(stage=stage, mission_id=mission_id, plan=plan,
                        ignore_dependencies=True, ignore_dependants=True))
    log.info(f"Started a new mission and triggered stage '{stage}' with all other stages ignored.")
    return True  # end


@init_client
def save(plan: Union[str, dict], client: Houston, **kwargs) -> bool:
    """Save a plan or update an existing plan.
    """
    client.delete_plan(safe=True)
    client.save_plan()
    log.info(f"Saved Plan '{client.plan['name']}' 🚀")
    return True  # end


@init_client
def delete(plan: str, client: Houston, **kwargs) -> bool:
    """Delete a plan.
    """
    client.delete_plan(safe=True)
    log.info(f"Deleted Plan '{client.plan['name']}'.")
    return True  # end


@init_client
def skip(plan: str, client: Houston, stage: str, mission_id: str, stages: Union[str, List[str]] = None,
         ignore_dependencies: bool = False, ignore_dependants: bool = False,  **kwargs) -> bool:
    """Start and end a stage without doing anything, essentially skipping it.
    """
    stages = stages if stages is not None else stage
    if isinstance(stages, str):
        stages = [a.strip() for a in stages.split(",")]
    for s in stages:
        client.start_stage(stage_name=s, mission_id=mission_id, ignore_dependencies=ignore_dependencies)
        client.end_stage(stage_name=s, mission_id=mission_id, ignore_dependencies=ignore_dependants)
        log.info(f"Marked stage '{s}' as complete without running it.")
    return True  # end


@init_client
def fail(plan: str, client: Houston, mission_id: str, stages: Union[str, List[str]] = None,
         stage=None, **kwargs) -> bool:
    """Force a stage or stages to be marked as failed.
    """
    stages = stages if stages is not None else stage
    if isinstance(stages, str):
        stages = [a.strip() for a in stages.split(",")]
    for s in stages:
        try:
            client.fail_stage(s, mission_id)
        except HoustonException:
            log.warning(f"Failed to fail stage '{s}'. Stage may not exist.")
            pass

    log.info(f"Marked stages as failed: {', '.join(stages)}")
    return True  # end


@init_client
def wait(plan: str, client: Houston, stage: str, mission_id: str, wait_callback: Callable[..., bool], start_time: float,
         time_limit_seconds, wait_interval_seconds, wait_params: Optional[dict] = None,
         wait_invocation_count: int = 1, **kwargs) -> bool:
    """Continue a stage that has already started and use the `wait_callback` function provided to check if the stage
    has finished. If the `wait_callback` returns `True`, the stage will end.
    This command is used by Houston services when a running a stage that takes longer than the service's
    execution time limit. If the service is going to run out of time it will trigger itself and continue waiting in a
    new invocation.

    See the docs for examples of using this command:
      - [Cloud Function](https://callhouston.io/docs/) # TODO: link to docs for how to use

    The triggering message can contain `wait_params` which will be used as parameters for the `wait_callback`. This
    command will trigger new runs a maximum of 500 times to prevent infinite loops. This can be changed with the
    `HOUSTON_MAX_WAIT_INVOCATIONS` environment variable.

    :param plan: Houston plan name to which the stage belongs.
    :param client: Houston client.
    :param stage: Name of the stage of the plan to be waited for.
    :param mission_id: Houston mission that this stage is a part of.
    :param wait_callback: A function that returns true when the task is finished and false when waiting should continue.
    :param start_time: Time the function started.
    :param time_limit_seconds: The maximum amount of time the service should wait for.
    :param wait_interval_seconds: Time to wait between running the wait callback.
    :param wait_params: Parameters to be given to the wait callback function.
    :param wait_invocation_count: A counter to track how many times a service has run the wait command for this stage.
    """

    if wait_invocation_count > HOUSTON_MAX_WAIT_INVOCATIONS:
        log.error(f"There have been over {HOUSTON_MAX_WAIT_INVOCATIONS} invocations for waiting! "
                  f"This could be an infinite loop; waiting will stop.")
        return True  # end

    log.info(f"⏳ Waiting for {stage} to finish. Time limit is {time_limit_seconds}s.")
    log.info(f"This is wait invocation number {wait_invocation_count}.")
    seconds_elapsed = time.time() - start_time
    wait_res = False
    wrapped_wait_callback = retry_wrapper(wait_callback)

    if wait_params is None:
        wait_params = dict()

    params = prepare_params(wait_params, wait_callback,
                            dict(plan=plan, mission_id=mission_id, stage=stage, command="wait",
                                 wait_invocation_count=wait_invocation_count))

    try:
        # run the waiting function until it returns True or time runs out
        while seconds_elapsed < time_limit_seconds and not wait_res:
            log.info(f"⏳ Not finished after {seconds_elapsed} seconds.")
            time.sleep(wait_interval_seconds)
            wait_res = wrapped_wait_callback(**params)
            seconds_elapsed = time.time() - start_time

    except BaseException as be:
        log.error(f"Exception has occurred while waiting for stage '{stage}' to complete: {be}.")
        log.error(f"Marking stage as failed.")
        client.fail_stage(stage, mission_id)
        raise be

    if wait_res:
        try:
            res = client.end_stage(stage, mission_id)
            client.trigger_all(res['next'], mission_id)
        except HoustonClientError:
            # fail silently for 'This stage has already been completed' errors after waiting ends as they are harmless
            log.info("Stage has already been completed - doing nothing")

        log.info(f"🏁 finished after {seconds_elapsed} seconds!")
        return True  # end

    else:
        log.info(f"⌛ Reached Cloud Function time limit.️ "
                 f"Waiting for stage '{stage}' will continue in new invocation.")

        # start new function to continue waiting
        event = dict(plan=plan, stage=stage, mission_id=mission_id, command=wait,
                     wait_invocation_count=wait_invocation_count + 1)
        client.trigger(event)

        return True  # end


@init_client
def trigger(plan: str, client: Houston, mission_id: str, stages: Union[str, List[str]] = None, stage=None,
            ignore_dependencies: bool = False, ignore_dependants: bool = False, **kwargs) -> bool:
    """Manually trigger a stage or stages in an in-progress mission. This should only be us"""
    stages = stages if stages is not None else stage
    if isinstance(stages, str):
        stages = [a.strip() for a in stages.split(",")]
    for s in stages:
        client.trigger(dict(stage=s, mission_id=mission_id, ignore_dependencies=ignore_dependencies,
                            ignore_dependants=ignore_dependants))
    log.info(f"Triggered stages: {', '.join(stages)}")
    return True  # end


# aliases for some commands
update = save
blastoff = start
scrub = ignore
dummy = skip


def run_command(command_name: str, plan: str = None, client: Houston = None, *args, **kwargs) -> bool:
    """Select and run a command given the name of a command.
    """
    # map command names to functions
    command_map = dict(
        start=start,
        blastoff=start,
        missionstart=start,
        missionsequencestart=start,
        ignore=ignore,
        scrub=scrub,
        staticfire=static_fire,
        save=save,
        update=update,
        delete=delete,
        dummy=dummy,
        skip=skip,
        fail=fail,
        wait=wait,
        trigger=trigger,
    )

    command_name = command_name.replace("-", "").replace("_", "").replace("plan", "")

    if command_name in command_map:
        return command_map[command_name](plan, client, *args, **kwargs)
    else:
        raise ValueError(f"Houston Cloud Function does not recognise command '{command_name}'. "
                         f"Valid commands: {', '.join(command_map.keys())}")


#
# utils
#


def prepare_params(params: dict, target_func: Callable, houston_context) -> dict:
    """Prepare parameters from a stage triggering event for use in a service function. This includes parsing of
    JSON objects, removal of unexpected arguments, and addition of a logger or context arguments if required.

    :param params: Dictionary containing pairs of parameter names and values.
    :param target_func: The Houston service function.
    :param houston_context: Object containing information about the context that the stage is being run in, e.g. the
                            triggering event, which contains the stage, plan name, and mission id.
    :return: Dictionary containing pairs of prepared parameter names and values.
    """
    if params is None:
        params = dict()

    if isinstance(params, str):
        try:
            params = json.loads(params)
        except json.decoder.JSONDecodeError:
            raise ValueError("`params` could not be parsed. Should be a dict or valid JSON.")

    import inspect
    (args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, annotations) = inspect.getfullargspec(target_func)

    if varkw is None:  # if function does not accept extra key word arguments
        # remove any values that aren't used by the function to avoid 'got an unexpected keyword argument', e.g. 'topic'
        params = {key: value for key, value in params.items() if key in args}

    if varargs is not None:
        log.warning(
            f"Houston service cannot use a function that expects '*{varargs}' because argument order cannot be "
            f"guaranteed. Please use named parameters or '**kwargs'.")

    # try to JSON parse every string param
    for key in params:
        if isinstance(params[key], str):
            try:
                params[key] = json.loads(params[key])
            except json.decoder.JSONDecodeError:
                pass

    # if wrapped function expects a log parameter then add it
    if ('log' in args or varkw is not None) and 'log' not in params:
        params['log'] = log

    # if wrapped function expects a houston_context parameter then add it
    if ('houston_context' in args or varkw is not None) and 'houston_context' not in params:
        params['houston_context'] = houston_context

    return params




import os
from io import BytesIO
from google.cloud import storage


def download_file_as_text(gs_uri):
    client = storage.Client(project=os.getenv('GCP_PROJECT', os.getenv('PROJECT_ID')))

    try:
        gs_bucket, gs_path = gs_uri.replace("gs://", "").split("/", 1)
    except ValueError:
        raise ValueError(f"Couldn't download plan; '{gs_uri}' is not a valid Google Storage URI. "
                         f"URI should look like 'gs://my-bucket/path/to/my-plan.yaml")

    bucket = client.get_bucket(gs_bucket)
    blob = bucket.get_blob(gs_path)

    if blob is None:
        raise ValueError(f"No object exists at GCS location: {gs_uri} - cannot load plan.")

    file_buffer = BytesIO()
    blob.download_to_file(file_buffer)
    file_buffer.seek(0)

    return file_buffer.read().decode("utf-8")



"""Houston client exceptions"""


class HoustonException(Exception):
    """Base Houston client exception"""

    pass


class HoustonServerBusy(HoustonException):
    """Raised due to too many requests to process"""

    pass


class HoustonClientError(HoustonException):
    """Error raised when the client has made an invalid request"""

    pass


class HoustonServerError(HoustonException):
    """Error raised when Houston server has issues"""

    pass


class HoustonPlanNotFound(ValueError):
    """Error raised when Houston plan does not exist as a saved plan or local file"""

    pass




import base64
import json
import os
from houston.client import Houston
from houston.exceptions import HoustonServerError, HoustonServerBusy, HoustonPlanNotFound
from houston.gcp.secret_manager import get_secret
from houston.gcp.cloud_storage import download_file_as_text
from google.cloud import pubsub_v1
from google.api_core.exceptions import GoogleAPIError
from retry import retry

retry_wrapper = retry((HoustonServerError, HoustonServerBusy, OSError, GoogleAPIError), tries=3, delay=1, backoff=100)


@retry_wrapper
def pubsub_trigger(client: Houston, data: dict, topic=None):
    """Sends a message to the provided Pub/Sub topic with the provided data payload.
    :param client: Instance of the Houston client class
    :param data: content of the message to be sent. Should contain 'stage' and 'mission_id'. Can contain any
                      additional JSON serializable information.
    :param string topic: Google Pub/Sub topic name, e.g. 'topic-for-stage'. This can either be provided here or be
                         set as a parameter for the stage as 'topic' or 'psq'.
    """

    publisher_client = pubsub_v1.PublisherClient()
    if hasattr(client, 'project'):
        project = client.project
    else:
        project = os.getenv("GCP_PROJECT", os.getenv("PROJECT_ID", None))

    data, stage_params = client._validate_message_data(data)

    service = client.get_service_from_stage_name(data['stage'])

    if service is not None and service.get('trigger') is not None and service.get('trigger').get('project'):
        project = service.get('trigger').get('project')

    if project is None:
        raise ValueError(
            "Project is not set. Specify a 'project' in the service's trigger, "
            "or use GCPHouston.project = 'your-project-id', "
            "or set 'GCP_PROJECT' environment variable"
        )

    # try to find the topic name in the stage parameters
    if topic is None:
        if service is not None and 'trigger' in service:
            topic = service['trigger'].get('topic')
        elif stage_params:
            if stage_params['topic']:
                topic = stage_params['topic']
            elif stage_params['psq']:
                topic = stage_params['psq']

        if topic is None:
            raise ValueError("Pub/Sub topic name could not be determined. It can either be provided as an argument to "
                             "pubsub_trigger, or be a stage parameter with name 'topic' or 'psq'")

    future = publisher_client.publish(topic=f"projects/{project}/topics/{topic}", data=json.dumps(data).encode("utf-8"))
    future.result()


class GCPHouston(Houston):

    project = os.getenv("GCP_PROJECT", os.getenv("PROJECT_ID", None))
    topic = None

    @staticmethod
    def load_plan(path):
        if len(path) > 4 and path[:5] == "gs://":  # download plan from cloud storage
            try:
                return download_file_as_text(path)
            except ValueError:
                raise HoustonPlanNotFound(f"Could not find Houston plan at Google Storage URI: {path}.")
        else:
            return Houston.load_plan(path)

    def _find_api_key(self) -> str:
        """Attempt to load the Houston API key from the environment or from Google Cloud Secret Manager.

        :return:
        """
        api_key = Houston._find_api_key()

        if api_key is None:
            # attempt to find the API key stored in Google Secret Manager
            try:
                return get_secret(name=os.getenv('HOUSTON_KEY_SECRET_NAME', 'houston-key'), project=self.project)
            except ValueError:
                raise ValueError("Houston API key could not be found in 'HOUSTON_KEY' environment variable and could "
                                 "not be loaded from Google Cloud Secret Manager.")
            # TODO: See the docs for alternative ways of supplying the API key: https://callhouston.io/docs/  # TODO: create docs and link to them here
        return api_key

    def pubsub_trigger(self, data, topic=None):
        """Sends a message to the provided Pub/Sub topic with the provided data payload.

        :param dict data: content of the message to be sent. Should contain 'stage' and 'mission_id'. Can contain any
                          additional JSON serializable information.
        :param string topic: Google Pub/Sub topic name, e.g. 'topic-for-stage'. This can either be provided here or be
                             set as a parameter for the stage as 'topic' or 'psq'.
        """
        return pubsub_trigger(self, data, topic)

    @staticmethod
    def extract_stage_information(data) -> dict:
        """Static method to extract stage information from sent PubSub message"""
        e = json.loads(base64.b64decode(data))

        e['ignore_dependencies'] = e.get('ignore_dependencies', False)
        e['ignore_dependants'] = e.get('ignore_dependants', False)

        # allow 'mission' as an alias for 'mission_id'
        if 'mission' in e and 'mission_id' not in e:
            e['mission_id'] = e['mission']

        return e



import time
import unittest
import sys

if sys.version_info >= (3, 3):
    from unittest import mock
else:
    import mock
from houston.plugin.azure import AzureHouston
from tests.test_houston import MockResponse


class MockEventGridResponse:
    @staticmethod
    def publish_events(topic, events):
        return None


class TestEventGrid(unittest.TestCase):

    test_plan_description = {
        "name": "test",
        "stages": [
            {"name": "start", "downstream": ["end"], "params": {"table": "test.sql"}},
            {"name": "end", "params": {"table": "test.sql", "topic": "tp-test-end", "topic_key": "abc"}},  # note: it is not recommeneded to keep keys in stage params
        ],
    }

    def test_event_grid_trigger(self):
        with mock.patch(
            "houston.plugin.azure.EventGridClient"
        ) as event_grid_client:
            with mock.patch("houston.interface.requests.request") as http:
                http.return_value = MockResponse(
                    status_code=200,
                    json_data={
                        "success": True,
                        "complete": False,
                        "next": ["end"],
                        "params": {"end": {"table": "test.sql", "psq": "tp-test-end"}},
                    },
                )
                houston = AzureHouston(
                    api_key="test-key", plan=self.test_plan_description
                )
                response = houston.end_stage("start", "test-launch-id")
                event_grid_client.return_value = MockEventGridResponse

                for next_stage in response['next']:
                    params = houston.get_params(next_stage)
                    houston.event_grid_trigger({'stage': next_stage, 'mission_id': "test-launch-id"},
                                               topic=params['topic'], topic_key=params['topic_key'])


if __name__ == "__main__":
    unittest.main()



import json
import os
import logging
from collections import defaultdict
from typing import *
import urllib.parse
from retry import retry
from houston.exceptions import HoustonClientError, HoustonException, HoustonServerBusy, \
                               HoustonServerError, HoustonPlanNotFound
from houston.interface import InterfaceRequest


HOUSTON_BASE_URL = os.getenv("HOUSTON_BASE_URL", "https://callhouston.io/api/v1")

retry_wrapper = retry((HoustonServerError, HoustonServerBusy, OSError), tries=3, delay=1, backoff=100)
log = logging.getLogger(os.getenv('HOUSTON_LOG_NAME', "houston"))


class Houston:

    def __init__(self, plan: Union[dict, str], api_key: str = None):
        """
        :param plan: Can be either:
                        - string: name of an existing plan to be loaded
                        - dict: plan definition, see Plan docs for example
                        If a plan definition is provided, ensure it is loaded with load_plan to make it
                        available to other instances
        :param api_key: Api key provided from https://callhouston.io, see account information. Can also be set as
                        environment variable HOUSTON_KEY
        """
        if api_key is None:
            api_key = self._find_api_key()

        if api_key is None:
            raise ValueError("No API key was found. Provide 'api_key' parameter or set the 'HOUSTON_KEY' environment "
                             "variable.")

        self.key = api_key
        self.interface_request = InterfaceRequest(key=api_key)
        self.base_url = HOUSTON_BASE_URL

        # TODO: load mission instead of plan
        if isinstance(plan, str):
            try:
                self.plan = self.get_plan(plan)  # look for existing saved plan
            except HoustonClientError:
                self.plan = self.import_plan(plan)  # look for file containing plan
        else:
            self.plan = plan

        print(self.plan)

        if "name" not in self.plan:
            raise HoustonClientError(
                "Sorry, this plan is not formatted correctly - must contain a name"
            )

    @staticmethod
    def _find_api_key() -> str:
        """Attempt to load the Houston API key from the environment.
        """
        return os.getenv("HOUSTON_KEY", None)

    @staticmethod
    def load_plan(path) -> str:
        """Load plan from a file.
        """
        try:
            with open(path) as f:
                plan = f.read()
            return plan

        except FileNotFoundError:
            raise HoustonPlanNotFound(f"No existing plans or plan files were found for '{path}'.")

    @classmethod
    def import_plan(cls, path: str) -> dict:

        plan = cls.load_plan(path)

        if ".yaml" in path or ".yml" in path:
            import yaml
            try:
                plan = yaml.load(plan, Loader=yaml.SafeLoader)
            except yaml.YAMLError as e:
                log.error(f"Plan has .yaml file extension but is not valid YAML")
                raise e

        else:  # assume plan is json
            try:
                plan = json.loads(plan)
            except json.JSONDecodeError as e:
                log.error("Plan is not valid JSON")
                raise e

        # convert any stage parameters to json strings if they are objects
        if not isinstance(plan, str):
            if 'stages' in plan:
                for i, stage in enumerate(plan['stages']):
                    if 'params' in stage:
                        for param in stage['params']:
                            if isinstance(stage['params'][param], dict) or isinstance(stage['params'][param], list):
                                plan['stages'][i]['params'][param] = json.dumps(stage['params'][param])

        return plan

    @retry_wrapper
    def save_plan(self):
        """Sends selected plan to Houston server"""
        self.interface_request.request(
            method="post", uri=self.base_url + "/plans", data=json.dumps(self.plan)
        )

    @retry_wrapper
    def delete_plan(self, safe=False):
        """Removes selected plan from Houston server

        :param boolean safe: Ignore exception raised by invalid request i.e. plan doesn't exist, True = ignore
        """
        plan_name = self.plan["name"]
        self.interface_request.request(
            "delete", uri=self.base_url + "/plans/" + plan_name, safe=safe
        )

    @retry_wrapper
    def get_plan(self, plan_name=None):
        """Get saved plan detail from Houston server

        :param string plan_name: [Optional] name of plan to retrieve details of
        :return dict: plan detail
        """
        if plan_name is None:
            plan_name = self.plan["name"]

        status_code, response = self.interface_request.request(
            "get", uri=self.base_url + "/plans/" + plan_name
        )
        return response

    @retry_wrapper
    def create_mission(self):
        """Creates a new instance of a mission
        :return string: new mission instance id or False if error
        """
        status_code, response = self.interface_request.request(
            "post",
            uri=self.base_url + "/missions",
            data=json.dumps({"plan": self.plan["name"]}),
        )
        if "id" not in response:
            raise HoustonException(
                "Create mission operation did not return a mission_id, please retry"
            )

        return response["id"]

    @retry_wrapper
    def start_stage(self, stage_name, mission_id, retry=3, ignore_dependencies=False):
        """Starts Houston stage, returns current stage parameters if available

        :param string stage_name: name of stage to start
        :param string mission_id: unique identifier of mission currently being completed
        :param int retry: number of retries in the case of failures, (429 responses are common with complex graphs)
        default is 3
        :param bool ignore_dependencies: if set to True, the stage will be allowed to start regardless of the state of
                                         all upstream stages dependencies
        :returns dict: Houston response {"next": list(string), "complete": bool, "params": dict()}
        """

        payload = {"state": "started", "ignoreDependencies": ignore_dependencies}
        status_code, json_response = self.interface_request.request(
            "post",
            uri=self.base_url + "/missions/" + mission_id + "/stages/" + stage_name,
            data=json.dumps(payload),
            retry=retry,
        )

        return json_response

    @retry_wrapper
    def end_stage(self, stage_name, mission_id, retry=3, ignore_dependencies=False):
        """Ends a Houston stage, returns downstream stages and available parameters

        :param string stage_name: name of stage which has finished
        :param string mission_id: unique identifier of mission currently being completed
        :param int retry: number of retry attempts
        :param bool ignore_dependencies: if set to True, all downstream stages dependant on this stage will be ignored,
                                         effectively ending the mission early
        :returns dict: Houston response {"next": list(string), "complete": bool, "params": dict(stage_name: dict())}
                       params contains multiple stage's parameters stored by stage name as keys
                       (e.g. {"next": "stage-two", "complete": False, "params": {"stage-two": dict()}})
        """

        payload = {"state": "finished", "ignoreDependencies": ignore_dependencies}
        status_code, json_response = self.interface_request.request(
            "post",
            uri=self.base_url + "/missions/" + mission_id + "/stages/" + stage_name,
            data=json.dumps(payload),
            retry=retry,
        )

        return json_response

    @retry_wrapper
    def fail_stage(self, stage_name, mission_id, retry=3):
        """Marks a Houston stage as failed, which allows it to be started again, returns downstream stages and available
        parameters

        :param string stage_name: name of stage which has failed
        :param string mission_id: unique identifier of mission currently being completed
        :param int retry: number of retry attempts
        :returns dict: Houston response {"next": list(string), "complete": bool, "params": dict(stage_name: dict())}
                       params contains multiple stage's parameters stored by stage name as keys
                       (e.g. {"next": "stage-two", "complete": False, "params": {"stage-two": dict()}})
        """

        payload = {"state": "failed"}
        status_code, json_response = self.interface_request.request(
            "post",
            uri=self.base_url + "/missions/" + mission_id + "/stages/" + stage_name,
            data=json.dumps(payload),
            retry=retry,
        )

        return json_response

    @retry_wrapper
    def ignore_stage(self, stage_name, mission_id, retry=3):
        """Marks a Houston stage as ignored. See https://callhouston.io/docs#ignored for more information

        :param string stage_name: name of stage which should be ignored
        :param string mission_id: unique identifier of mission currently being completed
        :param int retry: number of retry attempts
        :returns dict: Houston response {"next": list(string), "complete": bool, "params": dict()}
        """

        payload = {"state": "ignored"}
        status_code, json_response = self.interface_request.request(
            "post",
            uri=self.base_url + "/missions/" + mission_id + "/stages/" + stage_name,
            data=json.dumps(payload),
            retry=retry,
        )

        return json_response

    @retry_wrapper
    def skip_stage(self, stage_name, mission_id, retry=3):
        """Marks a Houston stage as skipped, meaning the rest of the mission will continue as if that stage doesn't
        exist.

        :param string stage_name: name of stage which should be ignored
        :param string mission_id: unique identifier of mission currently being completed
        :param int retry: number of retry attempts
        :returns dict: Houston response {"next": list(string), "complete": bool, "params": dict()}
        """

        payload = {"state": "skipped"}
        status_code, json_response = self.interface_request.request(
            "post",
            uri=self.base_url + "/missions/" + mission_id + "/stages/" + stage_name,
            data=json.dumps(payload),
            retry=retry,
        )

        return json_response

    def get_stage(self, stage_name: str) -> Optional[dict]:
        """Returns the full definition of a stage within the plan. Returns `None` if the stage
        doesn't exist."""
        filtered_stages = [s for s in self.plan['stages'] if s['name'] == stage_name]

        if len(filtered_stages) < 1:
            return None
        elif len(filtered_stages) > 1:
            raise ValueError("Can't return params because more than one stage in the plan has the name '{}'. "
                             "Plan is not valid.".format(stage_name))

        return filtered_stages[0]

    def get_params(self, stage_name: str) -> Optional[dict]:
        """Returns the parameters for the provided stage name as defined in the plan. Returns `None` if the stage
        doesn't exist. Note: The plan used to retrieve parameters is the latest version and is not linked to any
        particular mission ID, it is therefore possible for parameter values returned from this method to differ from
        those belonging to the current mission.

        :param string stage_name: name of a stage in the plan
        :return collections.defaultdict: stage parameters as key value pairs
        """

        this_stage = self.get_stage(stage_name)
        if this_stage is None:
            return None

        if 'params' in this_stage:
            params = this_stage['params']

        else:
            params = dict()

        return defaultdict(lambda: None, params)

    def get_service_from_stage_name(self, stage_name: str) -> Optional[dict]:
        if 'services' not in self.plan:
            return None

        this_stage = self.get_stage(stage_name)
        if this_stage is None:
            return None

        if 'service' in this_stage:
            filtered_services = [s for s in self.plan['services'] if s['name'] == this_stage['service']]
            if len(filtered_services) < 1:
                return None
            return filtered_services[0]
        else:
            return None

    @property
    def independent_stages(self) -> List[dict]:
        """Find stages in a plan that have no dependencies and should therefore be triggered to start a mission.
        :return: List of all independent stages
        """
        has_dependencies = []
        for stage in self.plan['stages']:
            if 'upstream' in stage and stage['upstream'] is not None and len(stage['upstream']) > 0:
                has_dependencies.append(stage['name'])
            if 'downstream' in stage and stage['downstream'] is not None:
                if isinstance(stage['downstream'], str):
                    has_dependencies.append(stage['downstream'])
                else:
                    has_dependencies += stage['downstream']

        return [s for s in self.plan['stages'] if s['name'] not in has_dependencies]

    def _validate_message_data(self, data) -> (dict, dict):

        stage = data.get("stage")
        if stage is None:
            raise ValueError("Triggering message data does not contain 'stage'.")
        if not isinstance(stage, str):
            raise ValueError(f"Triggering message has an invalid value for 'stage'. Expected string, got '{stage}'")

        params = self.get_params(stage)
        if params is None:
            raise ValueError(f"Stage '{stage}' has no params. Cannot trigger a stage without the relevant params.")

        if 'plan' not in data:
            data['plan'] = self.plan['name']

        return data, params

    @retry_wrapper
    def http_trigger(self, data: dict):
        """Trigger a stage of the plan via HTTP GET request. The contents of `data` will be passed as URL query
        parameters. The service specified for the stage is expected to have a trigger with method 'http' and a value for
        'url'. The request will be made without waiting for the result to ensure that the process that makes the request
        can end without the next stage needing to finish.
        :param dict data: content of the message to be sent. Should contain 'stage' and 'mission_id'. Can contain any
                          additional JSON serializable information.
        """
        stage = data.get('stage')
        if stage is None:
            raise ValueError("Cannot trigger stage: A stage name was not provided.")

        service = self.get_service_from_stage_name(stage)

        if service is None:
            raise ValueError(f"Cannot trigger stage '{stage}': no service is defined for this stage. "
                             f"See https://callhouston.io/docs#services.")

        url = service.get('trigger').get('url')

        query = urllib.parse.urlencode(data, doseq=False)  # convert content to url query

        self.interface_request.request(
            "GET",
            uri=f"{url}?{query}",
            fire_and_forget=True,
        )

    def trigger(self, data: dict):
        """Trigger a stage using the first available method based on the stage params.
        :param data: event data to be sent to each stage's service to trigger the stage. Should contain the name of the
                     stage and mission ID.
        """
        data, params = self._validate_message_data(data)

        # determine how to trigger stage
        service = self.get_service_from_stage_name(data['stage'])

        if service is not None:
            trigger_method = service['trigger']['method'].lower()
        elif 'topic' in params and 'topic_key' in params:
            trigger_method = 'azure/event-grid'
        elif any([p in params for p in ('psq', 'topic')]):
            trigger_method = 'google/pubsub'
        else:
            raise ValueError("Couldn't find a way to trigger the stage. Add the required information to the "
                             "stage definition. See docs: callhouston.io/docs#services")

        if trigger_method == 'google/pubsub' or trigger_method == 'pubsub':
            try:
                from houston.gcp import pubsub_trigger
                pubsub_trigger(self, data)
            except ImportError:
                raise ImportError(f"Cannot use Pub/Sub to trigger stage because GCP plugin is not installed. "
                                  f"Use: `pip install houston-client[gcp]`")
        elif trigger_method == 'azure/event-grid' or trigger_method == 'event-grid' or trigger_method == 'eventgrid':
            try:
                from houston.plugin.azure import event_grid_trigger
                event_grid_trigger(self, data)
            except ImportError:
                raise ImportError(f"Cannot use Event Grid trigger because Azure plugin is not installed."
                                  f"Use: `pip install houston-client[azure]`")

        elif trigger_method == "http":
            self.http_trigger(data)  # TODO: get endpoint from services as well

        else:
            raise ValueError(f"Trigger method '{trigger_method}' is not recognised. "
                             f"Use one of: http, google/pubsub, azure/event-grid.")

    def trigger_all(self, stages: List[str], mission_id: str,
                    ignore_dependencies: bool = False, ignore_dependants: bool = False, **kwargs):
        """Trigger multiple stages from a list of stage names. Triggering event data will be automatically generated.
        Any other keyword arguments provided will be sent to all stages.
        :param stages: list of stage names
        :param mission_id: unique identifier of mission currently being completed
        :param ignore_dependencies: If true, all stages will be triggered with instructions to ignore upstream dependencies
        :param ignore_dependants: If true, all stages will be triggered with instructions to upstream downstream stages
        """
        for stage in stages:
            self.trigger(dict(stage=stage, mission_id=mission_id, plan=self.plan['name'],
                              ignore_dependencies=ignore_dependencies, ignore_dependants=ignore_dependants, **kwargs))



"""High level API for services that carry out Houston tasks or commands.
"""

import os
import time
import json
from typing import Optional, Callable
from logging import getLogger, Logger

from houston.client import Houston
from houston.commands import run_command, wait, prepare_params
from houston.exceptions import HoustonClientError


def execute_service(
        event: dict,
        func: Callable,
        name: str = "unnamed",
        time_limit_seconds: int = 300,
        wait_callback: Optional[Callable[..., bool]] = None,
        wait_interval_seconds: int = 10,
        log: Optional[Logger] = None,
        client_cls: Callable = Houston,
):
    """Executes a Houston stage or Houston command based on the event provided.

    :param event:
    :param func:
    :param log: Logger to use
    :param name:
    :param time_limit_seconds:
    :param wait_callback:
    :param wait_interval_seconds:
    :param client_cls: The Houston client class to use, i.e. Houston, GCPHouston, or AzureHouston.
    :return:
    """
    start = time.time()  # start time of the service used by wait callback
    if log is None:
        log = getLogger(os.getenv('HOUSTON_LOG_NAME', "houston"))

    if 'plan' not in event:  # if not using Houston
        log.info(f"No plan specified; running without Houston.")

        params = json.loads(event['params']) if 'params' in event else event
        params = prepare_params(params, func, houston_context=None)

        func_res = func(**params)

        if wait_callback is not None:
            log.info("Wait callback is defined but it is not currently possible to wait indefinitely "
                     "without Houston - will wait as long as possible in this invocation.")
            params = prepare_params(func_res, wait_callback, event)
            while not wait_callback(**params):
                log.info("Wait callback returned False. Waiting will continue.")
                time.sleep(10)
            log.info("Wait callback returned True. Waiting finished.")

        log.info(f"Finished {name}.")
        return func_res

    log.info(f"Initialising Houston client for plan '{event['plan']}'.")

    h = client_cls(plan=event['plan'])

    #
    # houston commands
    #

    if 'command' in event:
        event_params = dict(**event)
        command = event_params.pop('command')
        plan = event_params.pop('plan')
        log.info(f"Executing command '{command}'.")
        if run_command(command, plan=plan, client=h, **event_params, wait_callback=wait_callback,
                       start_time=start, time_limit_seconds=time_limit_seconds,
                       wait_interval_seconds=wait_interval_seconds):
            return  # end the cloud function if required for the command

    #
    # start stage
    #

    try:
        h.start_stage(event['stage'], event['mission_id'], ignore_dependencies=event["ignore_dependencies"])
    except HoustonClientError:
        log.info("Stage has already started - stopping")
        return

    log.info(f"houston stage '{event['stage']}' started successfully")

    #
    # run operation
    #

    params = prepare_params(h.get_params(event['stage']), func, houston_context=event)
    log.info(f"Loaded stage params: {params}")

    try:
        func_res = func(**params)

    except BaseException:
        log.error(f"Exception has occurred in stage '{event['stage']}' within in {name} Cloud Function.")
        log.error(f"Marking stage as failed.")
        h.fail_stage(event['stage'], event['mission_id'])
        raise

    if wait_callback is not None:  # check if waiting is required
        event['wait_params'] = func_res
        wait(event, h, wait_callback=wait_callback, start_time=start,
             time_limit_seconds=time_limit_seconds, wait_interval_seconds=wait_interval_seconds)
        return func_res  # end

    #
    # end stage
    #

    res = h.end_stage(event['stage'], mission_id=event['mission_id'], ignore_dependencies=event['ignore_dependants'])
    h.trigger_all(res['next'], mission_id=event['mission_id'])
    log.info(f"Finished {name}.")

    return func_res




import logging
import sys
from .client import Houston
from .commands import start, update, save, ignore, delete, fail, trigger, skip, static_fire

# default logger also logs to stdout
# set 'LOG_NAME' env var to use a different logger
log = logging.getLogger("houston")
log.addHandler(logging.StreamHandler(stream=sys.stdout))
log.setLevel(logging.INFO)




import argparse
from .commands import start, save, ignore, delete, fail, trigger, static_fire, skip, run_command

parser = argparse.ArgumentParser(prog="houston",
                                 description='Houston Command Line Tool. Executes all high-level commands. '
                                             'Full documentation available at https://callhouston.io/docs.')

commands = parser.add_subparsers(dest='command')

parser_start = commands.add_parser('start', description=start.__doc__)
parser_start.add_argument('-p', '--plan', type=str, help='Plan name', required=True)
parser_start.add_argument('-s', '--stage', type=str,
                          help='(optional) Comma separated list of stage names from which to start the '
                               'mission. Defaults to all stages with no upstream dependencies')
parser_start.add_argument('-i', '--ignore', type=str,
                          help='(optional) Comma separated list of stage names that should be ignored for the mission')

parser_save = commands.add_parser('save', description=save.__doc__)
parser_save.add_argument('-p', '--plan', type=str, required=True,
                         help='Plan file name, either local file path or Google Cloud Storage URI. '
                              'Plan can be either JSON or YAML')

parser_delete = commands.add_parser('delete', description=delete.__doc__)
parser_delete.add_argument('-p', '--plan', type=str, required=True, help='Plan name')

parser_skip = commands.add_parser('skip', description=skip.__doc__)
parser_skip.add_argument('-p', '--plan', type=str, required=True, help='Plan name')
parser_skip.add_argument('-m', '--mission_id', type=str, required=True, help='Mission ID')
parser_skip.add_argument('-s', '--stage', type=str, required=True,
                         help='Comma separated list of stage names to be marked as completed without running')
parser_skip.add_argument('-iup', '--ignore-dependencies', dest="ignore_dependencies", type=bool,
                         help='If true, ignore upstream stages', default=False)
parser_skip.add_argument('-idown', '--ignore-dependants', dest="ignore_dependants", type=bool,
                         help='If true, ignore downstream stages', default=False)

parser_ignore = commands.add_parser('ignore', description=ignore.__doc__)
parser_ignore.add_argument('-p', '--plan', type=str, required=True, help='Plan name')
parser_ignore.add_argument('-m', '--mission_id', type=str, required=True, help='Mission ID')
parser_ignore.add_argument('-s', '--stage', type=str,
                           help='(optional) Comma separated list of stage names to be ignored. Defaults to all stages')

parser_fail = commands.add_parser('fail', description=fail.__doc__)
parser_fail.add_argument('-p', '--plan', type=str, required=True, help='Plan name')
parser_fail.add_argument('-m', '--mission_id', type=str, required=True, help='Mission ID')
parser_fail.add_argument('-s', '--stage', type=str, required=True,
                         help='Comma separated list of stage names to be marked as failed')

parser_trigger = commands.add_parser('trigger', description=trigger.__doc__)
parser_trigger.add_argument('-p', '--plan', type=str, required=True, help='Plan name')
parser_trigger.add_argument('-m', '--mission_id', type=str, required=True, help='Mission ID')
parser_trigger.add_argument('-s', '--stage', type=str, required=True,
                            help='Comma separated list of stage names to be triggered')
parser_trigger.add_argument('-iup', '--ignore-dependencies', dest="ignore_dependencies", type=bool,
                            help='If true, ignore upstream stages', default=False)
parser_trigger.add_argument('-idown', '--ignore-dependants', dest="ignore_dependants", type=bool,
                            help='If true, ignore downstream stages', default=False)

parser_static_fire = commands.add_parser('static-fire', description=static_fire.__doc__)
parser_static_fire.add_argument('-p', '--plan', type=str, required=True, help='Plan name')
parser_static_fire.add_argument('-s', '--stage', type=str, required=True, help='Name of the stage to be triggered')


args = vars(parser.parse_args())

if args['command'] is not None:
    run_command(args['command'], **args)

else:
    parser.print_help()



"""Houston Utilities for Google Cloud Platform

PubSub utils:
Allows user to create Google Cloud Pub/Sub message according to the plan stage options, e.g.:

    h.project = "my-project-1234"  # set the Google Cloud project name in the client

    res = h.end_stage("load-data", mission_id)

    for next_stage in res['next']:

        h.pubsub_trigger({'stage': next_stage, 'mission_id': mission_id}, topic=h.get_params(next_stage)['topic'])

--> sends a Pub/Sub message to the next tasks' topics. This assumes we have given each stage a 'topic' parameter.

Note: The topic name can either be provided as an argument or can be set as a parameter for the stage as 'topic' or
'psq', in which case it will be found automatically.

or:

    h.project = "my-project-1234"  # set the Google Cloud project name in the client

    response = h.end_stage("load-data", mission_id)

    h.call_stage_via_pubsub(response, mission_id)  # assumes each stage has a 'psq' parameter which gives the topic name

"""

from houston.gcp.client import GCPHouston



import sys
import unittest

if sys.version_info >= (3, 3):
    from unittest import mock
else:
    import mock
from houston.client import Houston
from houston.exceptions import HoustonClientError, HoustonServerError


class MockResponse:
    """Requests response mock"""

    def __init__(self, status_code, json_data):
        self.status_code = status_code
        self.json_data = json_data
        self.headers = {"Content-Type": "application/json"}

    def json(self):
        return self.json_data


class TestPlan(unittest.TestCase):
    test_plan_description = {"name": "test", "stages": [{"name": "start"}]}

    def test_get_plan_success(self):
        with mock.patch("houston.interface.requests.request") as http:
            http.return_value = MockResponse(
                status_code=200, json_data=self.test_plan_description
            )
            houston = Houston(api_key="test", plan=self.test_plan_description)
            response = houston.get_plan()
            self.assertEqual(response, self.test_plan_description)

    def test_get_plan_client_error(self):
        with mock.patch("houston.interface.requests.request") as http:
            http.return_value = MockResponse(status_code=400, json_data={"error": ""})
            houston = Houston(api_key="test", plan=self.test_plan_description)
            self.assertRaises(HoustonClientError, houston.get_plan)

    def test_get_plan_server_error(self):
        with mock.patch("houston.interface.requests.request") as http:
            http.return_value = MockResponse(status_code=500, json_data={"error": ""})
            houston = Houston(api_key="test", plan=self.test_plan_description)
            self.assertRaises(HoustonServerError, houston.get_plan)

    def test_post_save_plan(self):
        with mock.patch("houston.interface.requests.request") as http:
            http.return_value = MockResponse(
                status_code=200, json_data=self.test_plan_description
            )
            houston = Houston(api_key="test", plan=self.test_plan_description)
            houston.save_plan()

    def test_delete_plan(self):
        with mock.patch("houston.interface.requests.request") as http:
            http.return_value = MockResponse(status_code=200, json_data=None)
            houston = Houston(api_key="test", plan=self.test_plan_description)
            houston.delete_plan()

    def test_delete_plan_safe(self):
        with mock.patch("houston.interface.requests.request") as http:
            http.return_value = MockResponse(status_code=400, json_data=None)
            houston = Houston(api_key="test", plan=self.test_plan_description)
            houston.delete_plan(safe=True)


class TestStage(unittest.TestCase):
    test_plan_description = {
        "name": "test",
        "stages": [
            {"name": "start", "downstream": ["end"], "params": {"table": "test.sql"}},
            {"name": "end", "params": {"table": "test.sql"}},
        ],
    }

    def test_start_stage(self):
        with mock.patch("houston.interface.requests.request") as http:
            http.return_value = MockResponse(
                status_code=200,
                json_data={
                    "next": [],
                    "complete": False,
                    "params": {"table": "test.sql"},
                },
            )
            houston = Houston(api_key="test", plan=self.test_plan_description)
            houston.start_stage("test", "launch-id")

    def test_get_params(self):
        houston = Houston(api_key="test", plan=self.test_plan_description)

        params = houston.get_params("start")
        self.assertEqual(params['table'], "test.sql")
        self.assertIs(params['notaparam'], None)

        not_params = houston.get_params("notastage")
        self.assertIs(not_params, None)


if __name__ == "__main__":
    unittest.main()



"""Utilities to integrate a Google Cloud Function with Houston
"""

import os
import logging
from functools import wraps
from typing import Callable, Optional

from retry import retry
from google.api_core.exceptions import GoogleAPIError
from google.cloud import logging_v2 as cloud_logging
from google.cloud.logging_v2.resource import Resource

from houston.gcp.client import GCPHouston
from houston.service import execute_service

FUNCTION_NAME = os.getenv('FUNCTION_NAME', os.getenv('FUNCTION_TARGET', "houston-cloud-function"))

log = logging.getLogger(os.getenv('HOUSTON_LOG_NAME', "cloudfunctions.googleapis.com%2Fcloud-functions"))

# set up logging for Google Cloud Function
log_handler = cloud_logging.Client().get_default_handler(
    resource=Resource(type="cloud_function", labels={
        "function_name": FUNCTION_NAME,
        "region": os.getenv('FUNCTION_REGION'),
        "project_id": os.getenv('GCP_PROJECT')}))

log.setLevel(os.getenv('LOG_LEVEL', 'INFO'))
log.addHandler(log_handler)

retry_wrapper = retry((OSError, AttributeError, GoogleAPIError), tries=3, backoff=3, delay=3, logger=log)


def service(name: str = "unnamed", time_limit_seconds: int = os.getenv('FUNCTION_TIMEOUT_SEC', 300),
            wait_callback: Optional[Callable[..., bool]] = None, wait_interval_seconds: int = 10):
    """Wrapper to convert any Python function to a Houston service that can be executed by a Google Cloud Function
    using a Pub/Sub trigger. Example usage:

        # main.py

        from houston.gcp.cloud_function import service

        @service(name="My Service")
        def main(param_1, param_2):
            print("hello")

    The above function could be deployed with:

        gcloud functions deploy my-function-name --runtime python39 --trigger-topic my-function-topic \
          --source . --entry-point main --timeout 540 --set-env-vars GCP_PROJECT=my-project-id

    Note: this also requires a requirements.txt to be present, which must contain houston-client>2

    The function can then execute stages, provided the Pub/Sub topic name is provided in the stage params in your plan.
    A message like the following would trigger it to run a stage:

        {"plan": "my-plan", "stage": "my-stage", "mission_id": "a0234gyil344enbp"}

    The stage parameters will be provided to the wrapped function as they are defined in the plan. For example, the
    stage config for the above example function could look like the following:

        name: my-stage
        params:
          topic: my-function-topic
          param_1: "foo"
          param_2: 123

    Messages can also contain Houston commands, e.g. a message like the following would save/update a plan.

        {"plan": "gs://my-bucket/my-plan.yaml", "command": "save"}

    The following would start a new mission:

        {"plan": "my-plan", "command": "start"}

    For more examples and all other commands, see the docs:  # TODO: link to docs
    """

    log.info(f"Creating Houston service '{name}' within {FUNCTION_NAME}.")

    def outer(func):

        @wraps(func)
        def inner(event, context=None):

            # event may just be the message as a dict, as in the case of function testing events
            e = GCPHouston.extract_stage_information(event['data']) if 'data' in event else event

            log.info(f"Starting {name} Cloud Function with event data: {e}")

            res = execute_service(name=name,
                                  func=retry_wrapper(func),
                                  event=e,
                                  wait_callback=wait_callback,
                                  time_limit_seconds=time_limit_seconds,
                                  wait_interval_seconds=wait_interval_seconds,
                                  log=log,
                                  client_cls=GCPHouston)

            return res
        return inner
    return outer




from google.cloud import secretmanager
from google.api_core.exceptions import GoogleAPIError, PermissionDenied, NotFound
from retry import retry
import os

GCP_PROJECT = os.getenv('GCP_PROJECT', os.getenv('PROJECT_ID'))


@retry((OSError, AttributeError, GoogleAPIError), tries=3, delay=1, backoff=10)
def get_secret(name: str, version: str = "latest", project=GCP_PROJECT) -> str:

    if project is None:
        raise ValueError(f"Project is unknown. Please provide the GCP project ID with the 'GCP_PROJECT' or 'PROJECT_ID' "
                         f"environment variable.")

    client = secretmanager.SecretManagerServiceClient()
    name = client.secret_version_path(project=project, secret=name, secret_version=version)

    try:
        response = client.access_secret_version(name=name)
    except NotFound:
        raise ValueError(f"Secret '{name}' was not found.")
    except PermissionDenied:
        raise ValueError(f"Permission denied when trying to access secret '{name}'. "
                         f"Caller must have roles/secretmanager.secretAccessor to get secrets values.")

    return response.payload.data.decode('UTF-8')




from setuptools import setup, find_packages
from os import path

# Copy README.md to project description
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="houston-client",
    packages=find_packages(),
    description="Houston Python Client",
    long_description=long_description,
    long_description_content_type="text/markdown",
    version="1.2.0b4",
    url="https://github.com/datasparq-intelligent-products/houston-python",
    author="James Watkinson, Matt Simmons & Ivan Nedjalkov",
    license="MIT",
    author_email="james.watkinson@datasparq.ai",
    python_requires=">=3.7",
    install_requires=[
        "requests>=2.18.0,<3.0.0dev",
        "requests[security]>=2.18.0,<3.0.0dev",
        "PyYAML<=6",
        "retry>=0.9.2,<1",
    ],
    extras_require={
        "gcp": [
            "google-cloud-secret-manager>=2.5",
            "google-cloud-pubsub>=1.2",
            "google-cloud-storage>=1.35",
            "google-cloud-logging>=2",
        ],
        "azure": [
            "azure-eventgrid>=1.2.0",
            "azure-mgmt-eventgrid>=2.2.0",
        ]
    },
    keywords=["houston"],
)



"""Houston Client Http Request Wrapper"""

import time
from random import random
import logging
import requests
import os

from houston.exceptions import HoustonServerBusy, HoustonClientError, HoustonServerError

log = logging.getLogger(os.getenv('HOUSTON_LOG_NAME', "houston"))


class InterfaceRequest:
    """Houston client interface request class"""

    def __init__(self, key):
        self.headers = {"x-access-key": key, "Content-Type": "application/json"}

    def request(self, method, uri, params=None, data=None, retry=3, safe=False, fire_and_forget=False):
        """
        Request a Houston resource

        :param string method: Http method required for request (e.g. GET, POST, DELETE etc)
        :param string uri: Complete URL of request, including schema (e.g. https://)
        :param dict params: Parameters to be sent with request (will be json encoded)
        :param dict data: Parameters to be sent with request (will be form encoded)
        :param int retry: Number of retry's to attempt with request (only used by 429 server responses)
        :param bool safe: Do not raise errors in-case of client error
        :param bool fire_and_forget: If true, do not wait for a response
        :return: HTTP response code and response payload parsed as dict
        """

        timeout = None
        if fire_and_forget:
            timeout = 1

        try:
            response = requests.request(
                method, uri, headers=self.headers, params=params, data=data, timeout=timeout,
            )
        except requests.exceptions.ReadTimeout:
            if fire_and_forget:
                return 200, dict()
            else:
                raise
        except requests.exceptions.ConnectionError:
            raise HoustonServerError(
                f"Unable to connect to Houston API server at url: {uri}. Is your Houston server running?"
            )

        # retry if server busy - this can be common in a large workflow due to operations being immutable. 572 is the
        # Houston API's code for DagLockedError. 429 is 'Too Many Requests'.
        if response.status_code in (429, 572):
            if retry > 0:
                time.sleep(random())
                self.request(method, uri, params, data, retry - 1)
            else:
                raise HoustonServerBusy(
                    "received too many 429 responses from server, please reduce the number of requests"
                )

        if 400 <= response.status_code < 500 and not safe:
            err_msg = self._parse_error(response)
            raise HoustonClientError(
                "Unknown client error occurred. Please check request"
                if err_msg is None
                else err_msg
            )

        if 400 <= response.status_code < 500 and safe:
            return response.status_code, None

        if response.status_code >= 500:
            err_msg = self._parse_error(response)
            raise HoustonServerError(
                "Unknown server error occurred. If this persists please contact support"
                if err_msg is None
                else err_msg
            )

        try:
            json_data = response.json()
        except ValueError:
            json_data = None

        return response.status_code, json_data

    @staticmethod
    def _parse_error(response):
        """
        Parses any version of the API payload when the status code is != 200

        :param response: a response object
        :return: Error message
        """
        if response.headers.get("Content-Type") == "application/json":
            try:
                payload = response.json()
                if "msg" in payload:
                    return payload["msg"]
                elif "error" in payload and "message" in payload:
                    return payload["error"] + ". " + payload["message"]
                elif "error" in payload:
                    return payload["error"]
            except ValueError:
                # Generic Error
                return None
